//
// zxlog.c
//
// Copyright (c) 2002 Michael Ringgaard. All rights reserved.
//
// ZignX Medical Audit Trail Logger
//

#include "zxlog.h"

int zignxio;
int ctrlio;

//#define ALIGN(n, size) ((((n) + (size) - 1) / (size)) * (size))

hmodule_t shellmod;

#define ESC 27

#define ZXL_IDENTIFY          "ZXL\r"

#define ZIGNX_COMMAND_TIMEOUT 10000
#define MAX_DATA_LEN          4096

unsigned char zignx_response[MAX_DATA_LEN];
int zignx_response_index;
handle_t zignx_response_received;
int logging_suspended = 0;

unsigned char cmddata[MAX_DATA_LEN];
unsigned char rspdata[MAX_DATA_LEN];

//
// zignx_receiver_task
//

void __stdcall zignx_receiver_task(void *arg)
{
  int rc;
  unsigned char ch;

  while (1)
  {
    rc = read(zignxio, &ch, 1);
    if (rc > 0) 
    {
      if (ch == '\r')
	eset(zignx_response_received);
      else if (zignx_response_index < MAX_DATA_LEN)
        zignx_response[zignx_response_index++] = ch;
    }
  }
}

//
// clear_zignx_response
//

void clear_zignx_response()
{
  zignx_response_index = 0;
  ereset(zignx_response_received);
}

//
// wait_zignx_response
//

int wait_zignx_response(unsigned int timeout)
{
  return wait(zignx_response_received, timeout);
}

//
// read_command
//
// ESC <type (1)> <reqid (1)> <len (2)> <data (len)> <checksum (1)>
//

int read_command(struct packet *pkt, int maxlen)
{
  int rc;
  unsigned char b;
  unsigned char checksum;
  char *data;
  char *p;
  unsigned char buf[4];
  int n;
  int left;

  // Get packet start
  while (1)
  {
    rc = read(ctrlio, &b, 1);
    if (rc <= 0) return 0;

    if (b == ESC) break;
    if (b == '!') break;
    if (b == '\r') break;
    if (b >= 'a' && b <= 'z') break;
  }

  // Handle empty command
  if (b == '\r')
  {
    pkt->type = PACKET_ASCII;
    pkt->reqid = 0;
    pkt->len = 0;
    pkt->data[0] = 0;
  }

  // Handle ascii commands
  if (b != ESC)
  {
    pkt->type = PACKET_ASCII;
    pkt->reqid = 0;

    // Read until line feed
    p = data = pkt->data;
    *p++ = b;
    while (1)
    {
      // Test for buffer overflow
      if (p - data >= maxlen - 2) return -EBUF;
    
      // Read next command character
      rc = read(ctrlio, &b, 1);
      if (rc == 0) return -ETIMEOUT;
      if (rc < 0) return rc;

      // Test for end of command
      if (b == '\r')
      {
	*p = 0;
	pkt->len = p - data;
	return 1;
      }

      // Insert in buffer
      *p++ = b;
    }
  }

  // Handle packets commands, read packet header
  checksum = 0;
  for (n = 0; n < 4; n++)
  {
    rc = read(ctrlio, &b, 1);
    if (rc == 0) return -ETIMEOUT;
    if (rc < 0) return rc;
    buf[n] = b;
    checksum += b;
  }
  
  // Extract fields from header
  pkt->type = buf[0];
  pkt->reqid = buf[1];
  pkt->len = buf[2] | (buf[3] << 8);

  // Check for buffer overflow
  if (pkt->len > maxlen) return -EBUF;

  // Read data
  p = data = pkt->data;
  left = pkt->len;
  while (left > 0)
  {
    rc = read(ctrlio, p, left);
    if (rc == 0) return -ETIMEOUT;
    if (rc < 0) return rc;

    for (n = 0; n < rc; n++) checksum += *p++;
  }

  // Read and verify checksum
  rc = read(ctrlio, &b, 1);
  if (rc == 0) return -ETIMEOUT;
  if (rc < 0) return rc;
  checksum += b;
  if (checksum != 0) return -EBADCRC;

  return 1;
}

//
// send_packet
//

void send_packet(struct packet *pkt)
{
  unsigned char checksum;
  unsigned char buf[5];
  int i;

  // Send header
  buf[0] = ESC;
  buf[1] = pkt->type;
  buf[2] = pkt->reqid;
  buf[3] = pkt->len & 0xFF;
  buf[4] = (pkt->len >> 8) & 0xFF;
  checksum = buf[1] + buf[2] + buf[3] + buf[4];
  write(ctrlio, buf, 5);

  // Send data
  write(ctrlio, pkt->data, pkt->len);

  // Compute and send checksum
  for (i = 0; i < pkt->len; i++) checksum += pkt->data[i];
  checksum = 0x100 - checksum;
  write(ctrlio, &checksum, 1);
}

//
// send_error
//

void send_error(int reqid, char *msg)
{
  struct packet err;

  err.type = PACKET_ERROR;
  err.reqid = reqid;
  err.len = strlen(msg + 1);
  err.data = (unsigned char *) msg;

  send_packet(&err);
  printf("Error: %s\n", msg);
}

//
// handle_zignx_command
//

void handle_zignx_command(struct packet *req)
{
  struct record *rec;
  int len;

  printf("zignx command: %s\n", req->data);

  // Generate log record
  rec = (struct record *) malloc(sizeof(struct record) + req->len);
  rec->type = RECTYPE_ZIGNXCMD;
  rec->timestamp = time();
  rec->len = req->len;
  memcpy(rec->data, req->data, req->len);
  queue_log_record(rec, INFINITE);

  // Send command to ZignX and wait for response
  clear_zignx_response();
  write(zignxio, req->data, req->len);
  write(zignxio, "\r", 1);

  if (wait_zignx_response(ZIGNX_COMMAND_TIMEOUT) < 0)
  {
    write(ctrlio, "Timeout\r", 8);
  }
  else
  {
    // Store response
    len = zignx_response_index;
    memcpy(rspdata, zignx_response, len);
    rspdata[len] = 0;
    printf("zignx response: %s\n", rspdata);

    // Log response
    rec = (struct record *) malloc(sizeof(struct record) + len);
    rec->type = RECTYPE_ZIGNXRSP;
    rec->timestamp = time();
    rec->len = len;
    memcpy(rec->data, rspdata, len);
    queue_log_record(rec, INFINITE);

    // Relay response
    write(ctrlio, rspdata, len);
    write(ctrlio, "\r", 1);
  }
}

//
// handle_shell_command
//

void handle_shell_command()
{
  int prev_in = gettib()->in;
  int prev_out = gettib()->out;
  int prev_err = gettib()->err;

  if (!shellmod) shellmod = load("sh.exe");
  if (shellmod) 
  {
    gettib()->in = ctrlio;
    gettib()->out = ctrlio;
    gettib()->err = ctrlio;

    exec(shellmod, "");

    gettib()->in = prev_in;
    gettib()->out = prev_out;
    gettib()->err = prev_err;
  }
}

//
// handle_ok_command
//

void handle_ok_command()
{
  printf("ok command\n");
  write(ctrlio, ZXL_IDENTIFY, strlen(ZXL_IDENTIFY));
}

//
// handle_read_log
//

void handle_read_log(struct packet *req)
{
  int blksize;
  int xfersize;
  int rc;
  struct packet pkt;
  unsigned char *buf;
  int pos;
  int bytes;
  int param;

  // Get parameters
  if (req->len != 4)
  {
    send_error(req->reqid, "Invalid request size");
    return;
  }

  blksize = *(int *) req->data;

  // Get parameters
  blksize &= ~(SECTORSIZE - 1);
  if (blksize <= 0 || blksize > M)
  {
    send_error(req->reqid, "Invalid block size");
    return;
  }

  // Suspend log
  if (!logging_suspended)
  {
    rc = suspend_log();
    if (rc < 0)
    {
      send_error(req->reqid, "Unable to suspend logging");
      return;
    }
    logging_suspended = 1;
  }

  // Allocate xfer buffer
  buf = (unsigned char *) malloc(blksize + 4);
  if (!buf)
  {
    send_error(req->reqid, "Unable to allocate transfer buffer");
    return;
  }

  // Get log transfer size
  xfersize = current_log_length();
  printf("Log xfer %d bytes\n");

  // Send xfer start
  pkt.type = PACKET_XFER_START;
  pkt.reqid = req->reqid;
  pkt.len = 4;
  pkt.data = (unsigned char *) &xfersize;
  send_packet(&pkt);

  // Send log data
  pos = 0;
  while (pos < xfersize)
  {
    // Read log data into transfer buffer
    bytes = blksize;
    if (blksize > xfersize - pos) bytes = xfersize - pos;
    rc = read_log(pos, buf + 4, bytes);
    if (rc < bytes)
    {
      send_error(req->reqid, "I/O read error");
      free(buf);
      return;
    }

    // Send data block
    pkt.type = PACKET_XFER_DATA;
    pkt.reqid = req->reqid;
    pkt.len = bytes + 4;
    pkt.data = buf;
    *(int *) buf = pos;
    send_packet(&pkt);

    // Get acknowledge
    pkt.data = (unsigned char *) &param;
    rc = read_command(&pkt, 4);
    if (rc <= 0)
    {
      printf("Error: read log timeout, log transfer aborted\n");
      free(buf);
      return;
    }

    if (pkt.type == PACKET_XFER_ACK)
    {
      if (param >= 0 && param <= xfersize) pos = param;
    }
    else if (pkt.type == PACKET_XFER_ABORT)
    {
      printf("log transfer aborted\n");
      free(buf);
      return;
    }
    else
    {
      printf("Error: illegal protocol response, log transfer aborted\n");
      free(buf);
      return;
    }
  }

  // Log transfer complete
  pkt.type = PACKET_XFER_END;
  pkt.reqid = req->reqid;
  pkt.len = 0;
  send_packet(&pkt);
  free(buf);
}

//
// handle_clear_log
//

void handle_clear_log(struct packet *req)
{
  struct packet rsp;

  if (logging_suspended)
  {
    resume_log();
    logging_suspended = 0;
  
    rsp.type = PACKET_OK;
    rsp.reqid = req->reqid;
    rsp.len = 0;
    send_packet(&rsp);
  }
  else
    send_error(req->reqid, "Unable to clear log");
}

//
// start_command
//

void start_command()
{
  int rc;
  struct packet pkt;

  // Start ZignX response receiver
  zignx_response_index = 0;
  zignx_response_received = mkevent(1, 0);
  beginthread(zignx_receiver_task, 0, NULL, 0, NULL);

  // Open control port to pc
  ctrlio = openport("control-port");
  if (ctrlio < 0) 
  {
    printf("Error %d opening control port\n", ctrlio);
    return;
  }

  // Ready to read commands from control port
  while (1)
  {
    pkt.data = cmddata;
    rc = read_command(&pkt, MAX_DATA_LEN);
    if (rc == 0) 
    {
      if (logging_suspended)
      {
	printf("logging resume due to log clear timeout\n");
        resume_log();
        logging_suspended = 0;
      }

      continue;
    }

    if (rc < 0)
    {
      printf("Error %d receiving control command\n", rc);
      continue;
    }

    printf("control: packet type=%d reqid=%d len=%d\n", pkt.type, pkt.reqid, pkt.len);

    // Handle commands
    if (pkt.type == PACKET_ASCII)
    {
      if (pkt.len == 0)
        write(ctrlio, "\r", 1);
      else if (strcmp(pkt.data, "!shell") == 0)
	handle_shell_command();
      else if (strcmp(pkt.data, "ok") == 0)
	handle_ok_command();
      else
        handle_zignx_command(&pkt);
    }
    else if (pkt.type == PACKET_LOG_READ)
      handle_read_log(&pkt);
    else if (pkt.type == PACKET_LOG_CLEAR)
      handle_clear_log(&pkt);
    else
      send_error(pkt.reqid, "Invalid command");
  }
}
