//
// zxfile.c
//
// Copyright (c) 2002 Michael Ringgaard. All rights reserved.
//
// ZignX Medical Audit Trail Logger
//

#include "zxlog.h"

#define CHECKPOINT_INTERVAL   5

#define FLUSH_TIMEOUT         5000
#define SUSPEND_TIMEOUT       10000

struct queue *log_queue;
int logdev;
int logcfg;

handle_t logging_suspended;
handle_t resume_logging;

int logsize;
int logstart;    // Log must always start on a sector boundary
int logend;

int last_logstart;
int last_logend;
time_t next_checkpoint;

char sect[SECTORSIZE];

char *eventname[] =
{
  "NULL",
  "START",
  "SHUTDOWN",
  "SYSTEM",
  "ZIGNX",
  "INSTRUMENT",
  "PC",
  "KEYBOARD",
  "ZIGNXCMD",
  "ZIGNXRSP",
  "AUX"
};

//
// read_log_cfg
//

int read_log_cfg()
{
  lseek(logcfg, 0, SEEK_SET);
  read(logcfg, &logstart, sizeof(logstart));
  read(logcfg, &logend, sizeof(logend));
  return 0;
}

//
// write_log_cfg
//

int write_log_cfg()
{
  lseek(logcfg, 0, SEEK_SET);
  write(logcfg, &logstart, sizeof(logstart));
  write(logcfg, &logend, sizeof(logend));
  flush(logcfg);
  return 0;
}

//
// queue_log_record
//

int queue_log_record(struct record *rec, unsigned int timeout)
{
  return enqueue(log_queue, rec, timeout);
}

//
// suspend_log
//

int suspend_log()
{
  struct record rec;

  memset(&rec, 0, sizeof(struct record));
  rec.type = RECTYPE_NULL;

  if (enqueue(log_queue, &rec, SUSPEND_TIMEOUT) < 0) return -ETIMEOUT;
  if (wait(logging_suspended, SUSPEND_TIMEOUT) < 0) return -ETIMEOUT;
  return 0;
}

//
// resume_log
//

int resume_log()
{
  return eset(resume_logging);
}

//
// current_log_length
//

int current_log_length()
{
  if (logend < logstart)
    return logstart - logend;
  else
    return logstart + logsize - logend;
}

//
// read_log
//

int read_log(int pos, unsigned char *buffer, int count)
{
  int start;
  int rc;

  if (count % SECTORSIZE) count += (SECTORSIZE - count % SECTORSIZE);

  start = (pos + logstart) % logsize;
  if (start + count > logsize)
  {
    lseek(logdev, start, SEEK_SET);
    rc = read(logdev, buffer, logsize - start);
    if (rc < 0) return rc;
    buffer += logsize - start;
    count -= logsize - start;
    start = 0;
  }

  lseek(logdev, start, SEEK_SET);
  rc = read(logdev, buffer, count);
  if (rc < 0) return rc;

  return count;
}

//
// clear_log
//

void clear_log()
{
  logstart = (logstart & ~(SECTORSIZE - 1)) + 1;
  if (logstart >= logsize) logstart = 0;
  logend = logstart;
}

//
// checkpoint
//

void checkpoint(int force)
{
  time_t now = time();

  if (force || now >= next_checkpoint)
  {
    if (last_logstart != logstart || last_logend != logend)
    {
      //printf("log: checkpoint %d %d\n", logstart, logend);
      write_log_cfg();
      last_logstart = logstart;
      last_logend = logend;
    }

    next_checkpoint = now + CHECKPOINT_INTERVAL;
  }
}

//
// log_task
//

void __stdcall log_task(void *arg)
{
  struct record *rec;
  int dirty;
  int sectno;
  int sectofs;
  char *bufp;
  int len;
  int full;

  // Get log size
  logsize = fstat(logdev, NULL);
  last_logstart = logstart;
  last_logend = logend;
  next_checkpoint = time();

  printf("Logger started (%d KB) start: %d end: %d\n", logsize / K, logstart, logend);

  // If log does not end on a sector boundary we need to read the last sector of the log
  if (logend % SECTORSIZE)
  {
    lseek(logdev, (logend / SECTORSIZE) * SECTORSIZE, SEEK_SET);
    read(logdev, sect, SECTORSIZE);
  }
  else
    memset(sect, 0, SECTORSIZE);

  dirty = 0;
  while (1)
  {
    rec = dequeue(log_queue, FLUSH_TIMEOUT);

    if (rec)
    {
      int n;
      printf("%c -> ", *eventname[rec->type]);
      for (n = 0; n < rec->len; n++)
      {
	int ch;

	ch = (unsigned char) (rec->data[n]);
	if (ch < ' ' || ch > 127)
	  printf("[%02X]", ch);
	else
	  printf("%c", ch);
      }
      printf("\n");
    }

    //if (rec)
    //  printf("log: record type %s len %d\n", eventname[rec->type], rec->len);
    //else
    //  printf("log: timeout in: %d out: %d\n", log_queue->in, log_queue->out);

    if (rec && rec->type != RECTYPE_NULL)
    {
      // Check for log full
      len = sizeof(struct record) + rec->len;
      bufp = (char *) rec;

      if (logend < logstart)
	full = logend + len > logstart;
      else
	full = logend + len > logstart + logsize;

      if (full)
      {
	printf("log: log full\n");
	free(rec);
      }
      else
      {
	// Write log record to log device
	while (len > 0)
	{
	  sectno = logend / SECTORSIZE;
	  sectofs = logend % SECTORSIZE;

	  // See if we can fill the current log sector
	  if (len + sectofs >= SECTORSIZE)
	  {
	    // Fill sector buffer and write it to the log device
	    memcpy(sect + sectofs, bufp, SECTORSIZE - sectofs);
	    lseek(logdev, sectno * SECTORSIZE, SEEK_SET);
	    //printf("log: write sector %d\n", sectno);
	    write(logdev, sect, SECTORSIZE);
	    dirty = 0;
	    bufp += SECTORSIZE - sectofs;
	    len -= SECTORSIZE - sectofs;
	    logend += SECTORSIZE - sectofs;
	    memset(sect, 0, SECTORSIZE);

	    checkpoint(0);
	  }
	  else
	  {
	    // Copy the rest of the record to the sector buffer, and mark it dirty
	    memcpy(sect + sectofs, bufp, len);
	    dirty = 1;
	    logend += len;
	    len = 0;
	  }

	  // Check for log wrap
	  if (logend >= logsize) logend -= logsize;
	}

	// Deallocate record
	free(rec);
      }
    }
    else
    {
      // Flush current sector
      if (dirty)
      {
        sectno = logend / SECTORSIZE;
	lseek(logdev, sectno * SECTORSIZE, SEEK_SET);
        //printf("log: flush sector %d\n", sectno);
	write(logdev, sect, SECTORSIZE);
	dirty = 0;
      }

      // Checkpoint
      if (!dirty) checkpoint(0);

      // Check for suspend
      if (rec && rec->type == RECTYPE_NULL)
      {
	checkpoint(1);
	printf("log: logging suspended\n");
	wait(resume_logging, INFINITE);
	printf("log: logging resumed\n");
	checkpoint(1);
      }
    }
  }
}

//
// start_log
//

int start_log()
{
  int queuesize;
  char *logdevname;

  logcfg = open("/etc/zxlog.cfg", O_CREAT);
  if (logcfg < 0) return logcfg;
  if (fstat(logcfg, NULL) > 0) 
    read_log_cfg();
  else
  {
    logstart = 0;
    logend = 0;
  }

  logging_suspended = mkevent(0, 0);
  resume_logging = mkevent(0, 0);

  queuesize = get_numeric_property(zxlcfg, "zxlog", "queuesize", 256);
  log_queue = alloc_queue(queuesize);
  if (!log_queue) panic("unable to allocate log file queue");

  logdevname = get_property(zxlcfg, "zxlog", "logdevice", "/dev/hd0b");
  logdev = open(logdevname, O_RDWR);
  if (logdev < 0) return logdev;
  beginthread(log_task, 0, NULL, 0, NULL);

  return 0;
}
