//
// zxlog.c
//
// Copyright (c) 2002 Michael Ringgaard. All rights reserved.
//
// ZignX Medical Audit Trail Logger
//

#include "zxlog.h"

#define ZXL_VERSION "0.1"

struct section *zxlcfg;

char zignx_version[32];
char zignx_id[32];
int zignx_lang = 0;
char username[16];
char profilename[16];
int profile;

//
// printf
//

void printf(const char *fmt,...)
{
  va_list args;
  char buffer[1024];

  va_start(args, fmt);
  vsprintf(buffer, fmt, args);
  va_end(args);
  
  write(1, buffer, strlen(buffer));
}

//
// next_token
//

char *next_token(char **buf)
{
  char *p;
  char *q;

  if (**buf == 0) return "";
  p = q = *buf;
  while (*q != 0 && *q != ',') q++;
  if (*q == ',') *q++ = 0;
  *buf = q;
  return p;
}

//
// xtoi
//

int xtoi(char *s)
{
  int num = 0;
  while (*s)
  {
    if (*s >= '0' && *s <= '9')
      num = num * 16 + (*s - '0');
    else if (*s >= 'a' && *s <= 'f')
      num = num * 16 + (*s - 'a' + 10);
    else if (*s >= 'A' && *s <= 'F')
      num = num * 16 + (*s - 'A' + 10);
    else
      return -1;

    s++;
  }

  return num;
}

//
// openport
//

int openport(char *sectname)
{
  struct section *portcfg;
  int port;
  int rc;
  char *devname;
  char *parity;
  struct serial_config cfg;

  portcfg = find_section(zxlcfg, sectname);
  devname = find_property(portcfg, "device");
  cfg.speed = atoi(find_property(portcfg, "speed"));
  cfg.databits = atoi(find_property(portcfg, "databits"));
  cfg.stopbits = atoi(find_property(portcfg, "stopbits"));
  cfg.rx_timeout = atoi(find_property(portcfg, "rxtimeout"));
  cfg.tx_timeout = atoi(find_property(portcfg, "txtimeout"));
  parity = find_property(portcfg, "parity");
  switch (*parity)
  {
    case 'n': case 'N': cfg.parity = PARITY_NONE; break;
    case 'e': case 'E': cfg.parity = PARITY_EVEN; break;
    case 'o': case 'O': cfg.parity = PARITY_ODD; break;
    case 'm': case 'M': cfg.parity = PARITY_MARK; break;
    case 's': case 'S': cfg.parity = PARITY_SPACE; break;
  }

  port = open(devname, O_RDWR);
  if (port < 0) return port;

  rc = ioctl(port, IOCTL_SERIAL_SETCONFIG, &cfg, sizeof(struct serial_config));
  if (rc < 0) return rc;

  return port;
}

//
// send_zignx_comand
//

int send_zignx_comand(char *cmd)
{
  int rc;

  //printf("ZignX command: %s\n", cmd);

  rc = write(zignxio, cmd, strlen(cmd));
  if (rc < 0) return rc;

  rc = write(zignxio, "\r", 1);
  if (rc < 0) return rc;

  return 0;
}

//
// get_zignx_response
//

int get_zignx_response(char *buffer, int size)
{
  int rc;
  char ch;
  char *bufp;

  bufp = buffer;
  while (1)
  {
    rc = read(zignxio, &ch, 1);
    if (rc < 0) 
    {
      printf("Error %d in reading ZignX response\n", rc);
      return rc;
    }

    if (rc == 0) 
    {
      printf("Timeout reading ZignX response\n");
      return -ETIMEOUT;
    }

    if (ch == '\r') break;

    if (bufp - buffer >= size) return -EBUF;
    *bufp++ = ch;
  }

  *bufp = 0;
  //printf("ZignX response: %s\n", buffer);
  return bufp - buffer;
}

//
// parse_interval
//

int parse_interval(char **token, struct zignx_interval *interval)
{
  char *s;

  s = next_token(token);
  if (*s == '+')
    interval->start_hour = WEEKDAY_ALL;
  else if (*s == 'i')
    interval->start_hour = WEEKDAY_NONE;
  else
    interval->start_hour = atoi(s);

  interval->start_min = atoi(next_token(token));
  interval->end_hour = atoi(next_token(token));
  interval->end_min = atoi(next_token(token));

  return 0;
}

//
// parse_card_data
//

int parse_card_data(char *token, struct zignx_card *card)
{
  int i;

  strcpy(card->name, next_token(&token));
  strcpy(card->pincode, next_token(&token));
  card->lang = atoi(next_token(&token));

  for (i = 0; i < 10; i++) 
  {
    strcpy(card->profile_pincode[i], next_token(&token));
  }

  for (i = 0; i < 7; i++) 
  {
    parse_interval(&token, &card->weekday[i]);
  }

  for (i = 0; i < 20; i++) 
  {
    strcpy(card->domain[i], next_token(&token));
  }

  return 0;
}

//
// parse_profile_data
//

int parse_profile_data(char *token, struct zignx_profile *profile)
{
  int i;

  strcpy(profile->name, next_token(&token));
  strcpy(profile->pincode, next_token(&token));

  for (i = 0; i < 7; i++) 
  {
    parse_interval(&token, &profile->weekday[i]);
  }

  profile->powerout = xtoi(next_token(&token));
  profile->options = xtoi(next_token(&token));

  return 0;
}

//
// sync_time
//

int sync_time()
{
  char resp[128];
  char *token;
  struct tm tm;
  struct timeval tv;
  int rc;
  int retries;

  // Get time from ZignX, retry three times
  for (retries = 0; retries < 3; retries++)
  {
    send_zignx_comand("lt");
    rc = get_zignx_response(resp, 128);

    if (rc > 0 && strcmp(resp, "Error") != 0)
    {
      // Parse response
      token = resp;
      tm.tm_year = atoi(next_token(&token)) * 100;
      tm.tm_year += atoi(next_token(&token));
      tm.tm_year -= 1900;

      tm.tm_mon = atoi(next_token(&token)) - 1;
      tm.tm_mday = atoi(next_token(&token));
      
      tm.tm_hour = atoi(next_token(&token));
      tm.tm_min = atoi(next_token(&token));
      tm.tm_sec = 0;

      printf("Time is %04d/%02d/%02d %02d:%02d:%02d\n", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);

      // Update system clock
      tv.tv_sec = mktime(&tm);
      tv.tv_usec = 0;
      settimeofday(&tv);

      return 0;
    }
  }

  return -EIO;
}

//
// zignx_login
//

int zignx_login()
{
  int rc;
  char resp[512];
  //struct zignx_card card;
  //struct zignx_profile profile;

  // Get unique id from ZignX
  send_zignx_comand("li");
  rc = get_zignx_response(resp, 32);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  strcpy(zignx_id, resp);

  // Get version from ZignX
  send_zignx_comand("lv");
  rc = get_zignx_response(resp, 32);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  strcpy(zignx_version, resp);

  // Get language info
  send_zignx_comand("ll");
  rc = get_zignx_response(resp, 512);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  zignx_lang = atoi(resp);

#if 0
  // Get card info
  send_zignx_comand("lc");
  rc = get_zignx_response(resp, 512);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  parse_card_data(resp, &card);
  strcpy(username, card.name);
  printf("User: %s\n", username);

  // Get profile info
  send_zignx_comand("lp0");
  rc = get_zignx_response(resp, 512);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  parse_profile_data(resp, &profile);
  strcpy(profilename, profile.name);
  printf("Profile: %s\n", profilename);
#endif

  // Get language info
  send_zignx_comand("ll");
  rc = get_zignx_response(resp, 512);
  if (rc < 0) return rc;
  if (strcmp(resp, "Error") == 0) return -EZIGNX;
  zignx_lang = atoi(resp);

  printf("ZignX version %s (ID %s) %s\n", zignx_version, zignx_id, zignx_lang == 0 ? "US" : "DK");
  return 0;
}

//
// log_startup
//

void log_startup()
{
  struct record *rec;

  rec = malloc(sizeof(struct record) + 256);
  rec->type = RECTYPE_STARTUP;
  rec->timestamp = time();
  sprintf(rec->data, "%s,%s,%s", username, profilename, "Unknown");
  rec->len = strlen(rec->data);

  queue_log_record(rec, INFINITE);
}

//
// copy_zignx_log
//

int copy_zignx_log()
{
  int rc;
  char resp[512];
  char *buf;
  struct record *rec;

  // Get log from ZignX
  send_zignx_comand("lr");

  while (1)
  {
    rc = get_zignx_response(resp, 512);
    if (rc < 0) break;

    buf = resp;
    if (*buf == '\n') buf++;
    if (!*buf) break;

    printf("log: [%s]\n", buf);

    // Crack card name and profile from last profile start
    if (strlen(buf) > 50 && memcmp(buf + 17, "Profile start    ", 17) == 0)
    {
      char *p;
      char *q;

      p = buf + 34;
      q = profilename;
      while (*p != 0 && *p != ' ') *q++ = *p++;
      *q = 0;

      p = buf + 51;
      q = username;
      while (*p != 0 && *p != ' ') *q++ = *p++;
      *q = 0;

      if (!*username) strcpy(username, "<ANON>");
    }

    rec = malloc(sizeof(struct record) + strlen(buf));
    rec->type = RECTYPE_ZIGNX;
    rec->timestamp = time();
    rec->len = strlen(buf);
    memcpy(rec->data, buf, rec->len); 

    queue_log_record(rec, INFINITE);
  }

  // Clear log from ZignX
  send_zignx_comand("sr");
  rc = get_zignx_response(resp, 512);
  if (rc < 0 || strcmp(resp, "OK") != 0) printf("Error clearing ZignX log\n");

  return 0;
}

//
// main
//

int __stdcall main(hmodule_t hmod, char *cmdline, void *env)
{
  int rc;

  // Output banner
  printf("ZignX Medical Audit Trail Logger, version %s (Build %s %s)\n", ZXL_VERSION, __DATE__, __TIME__);
  printf("Copyright (c) 2002 Michael Ringgaard. All rights reserved.\n\n");

  // Read configuration
  printf("Reading configuration\n");
  zxlcfg = read_properties("/etc/zxlog.ini");
  if (!zxlcfg)
  {
    printf("Error reading configuration file\n");
    return 2;
  }

  // Open serial port to ZignX
  printf("Opening ZignX port\n");
  zignxio = openport("zignx-port");
  if (zignxio < 0)
  {
    printf("Error %d opening ZignX port\n", zignxio);
    return 2;
  }

  // Syncronize time with zignx
  printf("Updating system clock from ZignX\n");
  rc = sync_time();
  if (rc < 0)
  {
    printf("Error %d synchronizing time\n", rc);
    return 2;
  }

  // Retrieve login info from ZignX
  printf("Retrieving login info from ZignX\n");
  rc = zignx_login();
  if (rc < 0)
  {
    printf("Error %d retrieving login info\n", rc);
    return 2;
  }

  // Start logging
  rc = start_log();
  if (rc < 0)
  {
    printf("Error %d starting log writer\n", rc);
    return 2;
  }

  // Copy log from ZignX
  strcpy(username, "<UNKNOWN>");
  strcpy(profilename, "<UNKNOWN>");
  printf("Transfering log from ZignX\n");
  rc = copy_zignx_log();
  if (rc < 0)
  {
    printf("Error %d transfering ZignX log\n", rc);
    return 2;
  }

  // Log startup record
  log_startup();
  printf("STARTUP: user: %s profile: %s\n", username, profilename);

  // Start data collectors
  rc = init_data_logger();
  if (rc < 0)
  {
    printf("Error %d starting data collectors\n", rc);
    return 2;
  }

  // Start receiving commands on control port
  start_command();
  return 0;
}
