//
// zxlog.h
//
// Copyright (c) 2002 Michael Ringgaard. All rights reserved.
//
// ZignX Medical Audit Trail Logger
//

#ifndef ZXLOG_H
#define ZXLOG_H

#include <os.h>
#include <os/kbd.h>
#include <string.h>
#include <stdlib.h>
#include <inifile.h>

//
// Custom error codes
//

#define EZIGNX            1024
#define EBADCRC           1025

//
// Log record types
//

#define RECTYPE_NULL       0    // Void record
#define RECTYPE_STARTUP    1    // Startup (username, profile, station, domain)
#define RECTYPE_SHUTDOWN   2    // Shutdown
#define RECTYPE_SYSTEM     3    // System event
#define RECTYPE_ZIGNX      4    // ZignX log record
#define RECTYPE_INSTRDATA  5    // Instrument-to-PC data
#define RECTYPE_PCDATA     6    // PC-to-instrument data
#define RECTYPE_KEYBDATA   7    // Keyboard input data
#define RECTYPE_ZIGNXCMD   8    // ZignX command
#define RECTYPE_ZIGNXRSP   9    // ZignX response
#define RECTYPE_AUX        10   // Informational log record

#define MAX_RECLEN 255

//
// Packet types
//

#define PACKET_OK          0x00

#define PACKET_LOG_READ    0x01
#define PACKET_XFER_START  0x02
#define PACKET_XFER_DATA   0x03
#define PACKET_XFER_ACK    0x04
#define PACKET_XFER_ABORT  0x05
#define PACKET_XFER_END    0x06
#define PACKET_LOG_CLEAR   0x07

#define PACKET_ERROR       0xFE

#define PACKET_ASCII       0xFF

//
// Modes
//

#define MODE_LOG         0       
#define MODE_ADMIN       1

//
// Language codes
//

#define ZXLANG_ENGLISH   0
#define ZXLANG_DANISH    1

//
// Power out flags
//

#define POWEROUT_FDD       0x01
#define POWEROUT_OUT5      0x02
#define POWEROUT_CD        0x04
#define POWEROUT_OUT6      0x08
#define POWEROUT_OUT7      0x10
#define POWEROUT_HDD1      0x20
#define POWEROUT_OUT8      0x40
#define POWEROUT_HDD2      0x80

//
// Progfile options flags
//

#define OPTION_RESET       0x01
#define OPTION_POWER       0x02
#define OPTION_CARDCODE    0x04
#define OPTION_NOCARD      0x08
#define OPTION_AUTOSTART   0x10
#define OPTION_WOL         0x20
#define OPTION_NET         0x40
#define OPTION_MODEM       0x80

//
// ZignX card data
//

#define WEEKDAY_ALL        0xFF
#define WEEKDAY_NONE       0xFE

typedef char pincode_t[12];
typedef char domain_t[16];

struct zignx_interval
{
  unsigned char start_hour;
  unsigned char start_min;
  unsigned char end_hour;
  unsigned char end_min;
};

struct zignx_card
{
  char name[16];
  pincode_t pincode;
  int lang;
  pincode_t profile_pincode[10];
  struct zignx_interval weekday[7];
  domain_t domain[20];
};

struct zignx_profile
{
  char name[16];
  pincode_t pincode;
  struct zignx_interval weekday[7];
  unsigned char powerout;
  unsigned char options;
};

//
// Log record format
//

struct record
{
  unsigned char type;
  unsigned char len;
  time_t timestamp;
  char data[0];
};

//
// Packet
//

struct packet
{
  unsigned char type;
  unsigned char reqid;
  int len;
  unsigned char *data;
};

//
// zxqueue.c
//

struct queue
{
  handle_t notempty;
  handle_t notfull;
  int size;
  int in;
  int out;
  void **elems;
};

struct queue *alloc_queue(int size);
void free_queue(struct queue *q);
int enqueue(struct queue *q, void *msg, unsigned int timeout);
void *dequeue(struct queue *q, unsigned int timeout);

//
// zxfile.c
//

#define SECTORSIZE    512

int start_log();
int queue_log_record(struct record *rec, unsigned int timeout);

int suspend_log();
int resume_log();

int current_log_length();
int read_log(int pos, unsigned char *buffer, int count);
void clear_log();

//
// zxlog.c
//

int openport(char *sectname);

//
// zxdata.c
//

struct data_device
{
  char *name;
  handle_t collector;
  handle_t dev;
  int rectype;
  time_t timestamp;
  int datalen;
  char data[MAX_RECLEN];
  char buffer[MAX_RECLEN];
};

int init_data_logger();

//
// zxcmd.c
//

extern int zignxio;
extern int ctrlio;

void start_command();

//
// Global variables
//

extern struct section *zxlcfg;

extern char zignx_version[32];
extern char zignx_id[32];
extern int zignx_lang;
extern char username[16];
extern char profilename[16];
extern int profile;

void printf(const char *fmt,...);

#endif
